//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Methods for the class "fileInfo" (see fid.hh for more).
// -------------------------------------------------------------------
// $Id: fid.cxx,v 1.2 2002/12/29 13:43:20 loreti Exp $
// -------------------------------------------------------------------

#include <algorithm>
#include <fstream>
#include <iostream>
#include <iterator>
#include <stdexcept>
#include <sstream>
#include <cstring>
#include "fid.hh"
#include "main.hh"
#include "unint.hh"

extern "C" {
  #include <errno.h>
  #include <fcntl.h>
}

using std::cout;
using std::ostringstream;
using std::runtime_error;
using std::strerror;

fileInfo::fileInfo(
  const std::string & name
) : name_(name), content_(0) {

  // Constructor

  statFile();
  content_ = new char[size_];

  int ifs;
  if ((ifs = open(name_.c_str(), O_RDONLY | O_NDELAY)) < 0) {
    ostringstream s;

    s << progName << ": " << strerror(errno) << " - file "
      <<  name_;
    throw runtime_error(s.str());

  } else {
    ssize_t ifsIn;

    if ((ifsIn = read(ifs, content_, size_)) < 0) {
      ostringstream s;

      s << progName << ": " << strerror(errno) << " - file "
        << name_;
      close(ifs);
      throw runtime_error(s.str());
    }
    close(ifs);

    if (ifsIn != size_) {
      ostringstream s;

      s << progName << ": File size doesn't match: stat="
        << size_ << ", read=" << ifsIn << " - file "
        << name_;
      throw runtime_error(s.str());
    }
  }

  if (debug) {
    cout << progName << "File " << name_ << " read\n"
         << "----------Start----------\n";
    std::copy(content_, content_+size_,
              std::ostream_iterator<char>(cout, ""));
    cout << "-----------End-----------" << std::endl;
  }
}

void fileInfo::statFile()
{

  // Retrieves file informations calling stat(2)

  struct stat statBuf;

  if (stat(name_.c_str(), &statBuf) < 0) {
    ostringstream s;

    s << progName << ": " << strerror(errno)
      << " - file " << name_;
    throw runtime_error(s.str());
  }

  if (!S_ISREG(mode_ = statBuf.st_mode)) {
    ostringstream s;

    s << progName << ": Not a regular file - "
      << name_;
    throw runtime_error(s.str());
  }

  uid_   = statBuf.st_uid;
  gid_   = statBuf.st_gid;
  size_  = statBuf.st_size;
  mtime_ = statBuf.st_mtime;
}

void fileInfo::update()
{

  // Restores the file as it was

  ostringstream s;
  int           ofs;

  if ((ofs = open(name_.c_str(), O_WRONLY | O_CREAT | O_TRUNC, mode_)) < 0) {
    s << strerror(errno) << " - file " << name_ << '\n';
    errOut(s.str());
    return;
  }

  if (fchown(ofs, uid_, gid_) < 0) {
    s << strerror(errno) << " - file " << name_ << '\n';
    errOut(s.str());
    close(ofs);
    return;
  }

  if (write(ofs, content_, size_) < 0) {
    s << strerror(errno) << " - file " << name_ << '\n';
    errOut(s.str());
    close(ofs);
    return;
  }

  close(ofs);

  s << "File " << name_ << " restored\n";
  infoOut(s.str());
}

void fileInfo::checkStat(
  int    fd,
  mode_t mode,
  uid_t  uid,
  gid_t  gid
) const {

  // Compares (and, if needed, restores) all the file data (except the
  // content)

  if (mode != mode_) {
    if (fchmod(fd, mode_) < 0) {
      ostringstream s;

      s << strerror(errno) << ": couldn't restore mode - file "
        << name_ << '\n';
      errOut(s.str());
    } else {
      ostringstream s;

      s << "File mode restored: file " << name_ << '\n';
      infoOut(s.str());
    }
  }

  if (uid != uid_   ||   gid != gid_) {
    if (fchown(fd, uid_, gid_) < 0) {
      ostringstream s;

      s << strerror(errno) << ": couldn't restore owner - file "
        << name_ << '\n';
      errOut(s.str());
    } else {
      ostringstream s;

      s << "File owner restored: file " << name_ << '\n';
      infoOut(s.str());
    }
  }
}
