//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// A "file watcher" program.  Stores in memory a copy of the content
// of the files given on the command line; and compares, at constant
// time intervals, the current content with the stored copy.  If some
// difference is found, the original content is restored (and a
// message written into /var/log/messages).
//
// - The watched files must be small, of course
// - Also file mode and owner are watched (and restored, if needed)
// -------------------------------------------------------------------
// $Id: main.cxx,v 1.8 2006/09/11 12:23:21 loreti Exp $
// -------------------------------------------------------------------

#include <algorithm>
#include <functional>
#include <fstream>
#include <iostream>
#include <iterator>
#include <list>
#include <sstream>
#include <climits>
#include <cstdlib>
#include "main.hh"
#include "fid.hh"
#include "unint.hh"

extern "C" {
  #include <errno.h>
  #include <getopt.h>
  #include <signal.h>
  #include <fcntl.h>
}

using std::cerr;
using std::cout;
using std::endl;
using std::flush;
using std::list;
using std::ostringstream;

// An helper function object, to delete instances of fileInfo ADT's

struct deletePointed : public std::unary_function<fileInfo *, void> {
  void operator() (fileInfo * p) { delete p; }
};

// Global variables

bool   debug = false;
char * progName;

// Local global variables

namespace {
  const unsigned long defaultSleepTime = 60;
  unsigned long       sleepTime        = defaultSleepTime;
  list<fileInfo *>    lfi;

  const char Version[] = "1.2";
  const char Date[]    = "06-Jun-2005";

  // Files to watch by default

  const char watch1[]     = "/etc/passwd";
  const char watch2[]     = "/etc/shadow";
  const char watchDebug[] = "/root/fwd/prova.txt";

  // Local procedures

  char *        baseName(char *);
  void          cleanup();
  void          processArgs(int, char * []);
  bool          processFile(const fileInfo *);
  void          processList();
  void          processOpts(int, char * []);
  sigFunction * sigHand(int);
  void          usage();
} // namespace

// -------------------------------------------------------------------
// -------------------------------main--------------------------------
// -------------------------------------------------------------------

int main(
  int    argc,
  char * argv[]
) {

  // Initialization phase:
  // - Make "progName" point to the unqualified program name
  // - Process the command line options
  // - Establish an handler for the interrupt signal
  // - Process command line arguments

  progName = baseName(argv[0]);

  processOpts(argc, argv);

  if ((mySignal(SIGINT, reinterpret_cast<sigFunction *>(sigHand)))
      == SIG_ERR) {
    cerr << progName << ": Couldn't establish SIGINT handler\n";
    return EXIT_FAILURE;
  }

  processArgs(argc, argv);

  if (debug) {
    cout << progName << ": Sleep time " << sleepTime
         << " seconds" << endl;
  } else {
    daemonize(progName);
    infoOut("Starting\n");
  }

  // Main loop: waits for "sleepTime", then checks the watched files.
  // The main loop runs forever (exit with CTRL-C or "kill -s SIGINT
  // <pid>".

  for (;;) {
    sleep(sleepTime);

    if (debug) {
      ostringstream s;

      time_t now = time(0);
      cout << progName << ": Waked up at "
           << asctime(localtime(&now)) << flush;
    }
    processList();
  }
}

namespace {

// -------------------------------------------------------------------
// -----------------------------baseName------------------------------
// -------------------------------------------------------------------

  char * baseName(
    char * pc
  ){

    // Strips the (eventual) path informations from argv[0], passed as
    // formal parameter; if no file name is given, returns an empty
    // string.

    char * p = std::strrchr(pc, '/');
    if (p == 0) return pc;
    return ++p;
  }

// -------------------------------------------------------------------
// ------------------------------cleanup------------------------------
// -------------------------------------------------------------------

  void cleanup()
  {

    // Frees all the allocated memory

    std::for_each(lfi.begin(), lfi.end(), deletePointed());
  }

// -------------------------------------------------------------------
// ----------------------------processArgs----------------------------
// -------------------------------------------------------------------

  void processArgs(
    int    argc,
    char * argv[]
  ) {

    // Process command line arguments as names of files to be watched:
    // - Finds the absolute path to the file name;
    // - Creates a fileInfo instance for the given file;
    // - Stores the fileInfo in the "lfi" list;
    // - If the list is empty, defaults to "/etc/passwd" and
    //   "/etc/shadow" (or "/home/loreti/fwd/prova.txt" for
    //   debugging purposes).

    try {
      while (optind < argc) {
        char absPath[PATH_MAX];
        realpath(argv[optind++], absPath);
        fileInfo * pfi = new fileInfo(absPath);

        lfi.push_back(pfi);

        if (debug) {
          cout << progName << ": Watching file \""
                    << pfi->getName() << "\"\n";
          cout << "\tLast modification: "
               << asctime(localtime(&pfi->getMtime()));
          cout << "\tFile size: " << pfi->getSize() << " bytes"
               << endl;
        }
      }
    } catch (std::exception & e) {
      cerr << "* Exception * " << e.what() << '\n';
      cleanup();
      exit(EXIT_FAILURE);
    }

    if (lfi.empty()) {
      if (debug) {
        if (access(watchDebug, F_OK) == 0) {
          lfi.push_back( new fileInfo(watchDebug) );
        }
      } else {
        if (access(watch1, F_OK) == 0) {
          lfi.push_back( new fileInfo(watch1) );
        }
        if (access(watch2, F_OK) == 0) {
          lfi.push_back( new fileInfo(watch2) );
        }
      }
    }
  }

// -------------------------------------------------------------------
// ----------------------------processFile----------------------------
// -------------------------------------------------------------------

  bool processFile(
    const fileInfo * pfi
  ) {

    // Checks the file for which the initial informations were stored
    // in the fileInfo instance pointed to by "pfi".

    struct stat         statBuf;
    const std::string & fileName = pfi->getName();

    if (stat(fileName.c_str(), &statBuf) < 0) {
      ostringstream s;

      s << std::strerror(errno) << " - stat(" << fileName
        << ")\n";
      errOut(s.str());
      return true;
    }

    if (statBuf.st_size != pfi->getSize()) {
      ostringstream s;

      s << "File " << fileName << ": size has changed\n";
      errOut(s.str());
      return true;
    }

    int ifs;
    if ((ifs = open(fileName.c_str(), O_RDONLY | O_NDELAY)) < 0) {
      ostringstream s;

      s << "Couldn't open file " << fileName << '\n';
      errOut(s.str());
      return true;
    }

    off_t   size = pfi->getSize();
    char *  pc   = new char[ size ];
    ssize_t ifsIn;

    if ((ifsIn = read(ifs, pc, pfi->getSize())) < 0) {
      ostringstream s;

      s << "Couldn't read file " << fileName << '\n';
      errOut(s.str());
      return true;
    }

    bool retVal = ! std::equal(pc, pc+size, pfi->getContent());

    delete [] pc;

    if (retVal) {
      ostringstream s;

      s << "File " << fileName << ": content has changed\n";
      errOut(s.str());

    } else {
      pfi->checkStat(ifs, statBuf.st_mode, statBuf.st_uid,
                     statBuf.st_gid);
    }

    close(ifs);
    return retVal;
  }

// -------------------------------------------------------------------
// ----------------------------processList----------------------------
// -------------------------------------------------------------------

  void processList()
  {

    // Watches all the files in the list "lfi"

    list<fileInfo *>::const_iterator it, iEnd = lfi.end();

    for (it = lfi.begin();   it != iEnd;   ++it) {
      if (processFile(*it)) (*it)->update();
    }
  }

// -------------------------------------------------------------------
// ----------------------------processOpts----------------------------
// -------------------------------------------------------------------

  void processOpts(
    int    argc,
    char * argv[]
  ) {

    // Process the command options, using "getopt_long"

    char shortOpts[] = "ds:hv";

    struct option longOpts[] = {
      {"debug",      no_argument,       0, 'd'},
      {"sleep-time", required_argument, 0, 's'},
      {"help",       no_argument,       0, 'h'},
      {"version",    no_argument,       0, 'v'},
      {0,            0,                 0,  0 }
    };

    int c;
    while ((c = getopt_long(argc, argv, shortOpts, longOpts, 0)) != -1) {
      switch (c) {
        char * pc;

        case 'd':
          debug = true;
          break;

        case 's':
          sleepTime = std::strtoul(optarg, &pc, 10);

          if (*pc != '\0') {
            cerr << progName << ": Invalid sleep-time value \""
                 << optarg << "\"\n";
            usage();
            exit(EXIT_FAILURE);
          }
          break;

        case 'h':
          usage();
          exit(EXIT_SUCCESS);

        case 'v':
          cerr << progName << " version " << Version << " ("
               << Date << ")\n";
          exit(EXIT_SUCCESS);
      }
    }
  }

// -------------------------------------------------------------------
// ------------------------------sigHand------------------------------
// -------------------------------------------------------------------

  sigFunction * sigHand(
    const int signalNr
  ) {

    // Signal handler

    ostringstream s;

    switch (signalNr) {
      case SIGINT:
        s << "Interrupt (SIGINT or CTRL-C)";
        break;

      default:
        s << "Unknown signal = " << signalNr;
        break;
    }
    s << " - aborting\n";
    errOut(s.str());

    cleanup();
    exit(EXIT_SUCCESS);
  }

// -------------------------------------------------------------------
// ------------------------------usage--------------------------------
// -------------------------------------------------------------------

  void usage()
  {

    // Short summary of the program invocation rules

    cerr << "Usage:\t " << progName
         << " [options] file [ file [ file ... ] ] .\n";
    cerr << "Options: [-d | --debug]          : run interactively, "
      "debug printout\n";
    cerr << "\t [-s | --sleep-time] s : set sleep time in seconds "
      "(default " << defaultSleepTime << ")\n";
    cerr << "\t [-h | --help]           : this help\n";
    cerr << "\t [-v | --version]        : prints the version "
      "informations and exits\n";
  }

} // namespace
