//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Unix interface for the file watcher
// -------------------------------------------------------------------
// $Id: unint.cxx,v 1.2 2002/12/27 11:00:36 loreti Exp $
// -------------------------------------------------------------------

#include <iostream>
#include <sstream>
#include <stdexcept>
#include "main.hh"
#include "unint.hh"

extern "C" {
  #include <errno.h>
  #include <signal.h>
  #include <syslog.h>
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <unistd.h>
  #include <fcntl.h>
}

using std::ostringstream;
using std::runtime_error;
using std::strerror;
using std::string;

static const int OPEN_MAX_GUESS = 128;

void daemonize(
  const char * ident
) {

  // See W. Richard Stevens, "Advanced Programming in the UNIX
  // Environment", chapter 13 (pages 415-425) for details.
  //
  // - Forks and terminates the original task;
  // - becomes the session (and group) leader of a new session
  //   (and process group), and removes the controlling terminal;
  // - clears the file mode creation mask;
  // - forks and terminates the original task;
  // - closes all file descriptors;
  // - connects stdin, stdout and stderr to /dev/null;
  // - and finally opens a connection to /var/log/messages.

  int   openMax;
  pid_t pid;
  int   fd;

  if ((pid = fork()) < 0) {
    ostringstream s;

    s << ident << ": " << strerror(errno) << " - fork(1)\n";
    throw runtime_error(s.str());

  } else if (pid) {
    exit(EXIT_SUCCESS);         // Parent; bye bye...
  }

  setsid();
  chdir("/");
  umask(0);

  // The second fork prevents the acquisition of a controlling
  // terminal by guaranteeing that the process is not a session
  // leader.  Under BSD that happens only when an explicit ioctl
  // request is made, but under SVR4 it is automatic if a terminal
  // device is opened by a session leader and O_NOCTTY is not
  // specified.

  if ((pid = fork()) < 0) {
    ostringstream s;

    s << ident << ": " << strerror(errno) << " - fork(2)\n";
    throw runtime_error(s.str());

  } else if (pid) {             // Parent
    exit(EXIT_SUCCESS);
  }

  // In a running daemon, all file descriptors must be closed.  Other
  // unices have OPEN_MAX; under Linux we have to call "sysconf" to
  // know the maximum number of allowed open files.

  if ((openMax = sysconf(_SC_OPEN_MAX)) < 0) {
    ostringstream s;

    s << ident << ": " << strerror(errno) << " - sysconf\n";
    throw runtime_error(s.str());

  } else if (openMax == 0) {
    openMax = OPEN_MAX_GUESS;   // Queried option not available
  }

  for (fd = 0;   fd < openMax;   ++fd) {
    close(fd);
  }

  if ((fd = open("/dev/null", O_RDWR)) < 0) {
    exit(EXIT_FAILURE);
  }

  if (fd != STDIN_FILENO            ||
      dup2(fd, STDOUT_FILENO) < 0   ||
      dup2(fd, STDERR_FILENO) < 0)
    {
      exit(EXIT_FAILURE);
    }

  openlog(ident, LOG_PID | LOG_CONS, LOG_DAEMON);
}

sigFunction * mySignal(
  const int     sigNr,
  sigFunction * fun
) {
  struct sigaction action, oldAction;

  //  - Declare "fun" as handler for "sigNr": no additional signals
  //    to be blocked when our signal handler is called.

  sigemptyset(&action.sa_mask);
  action.sa_handler = fun;
  action.sa_flags   = 0;

  if (sigNr == SIGALRM) {
    action.sa_flags |= SA_INTERRUPT;
  } else {
    action.sa_flags |= SA_RESTART;
  }

  if (sigaction(sigNr, &action, &oldAction) < 0) {
    return SIG_ERR;
  } else {
    return oldAction.sa_handler;
  }
}

void errOut(
  const string & s
) {

  // Error message, to standard error or to /var/log/messages

  if (debug) {
    std::cerr << s;
  } else {
    syslog(LOG_ERR | LOG_DAEMON, s.c_str());
  }
}

void infoOut(
  const string & s
) {

  // Error message, to standard output or to /var/log/messages

  if (debug) {
    std::cout << s << std::flush;
  } else {
    syslog(LOG_INFO | LOG_DAEMON, s.c_str());
  }
}
